import logging
import pytest

LOGGER = logging.getLogger(__name__)


@pytest.fixture(scope='function')
def example_fixture():
    LOGGER.info("Setting Up Example Fixture...")
    yield
    LOGGER.info("Tearing Down Example Fixture...")



import abc
from typing import Optional, List, Dict

from .model import TimeEvent, TimeContext, Period, Label


class AbstractTimeRepository(abc.ABC):
    name: str

    def register(self, time_context: TimeContext):
        raise NotImplementedError

    def add(self, time_event: TimeEvent):
        raise NotImplementedError

    def get(self, event_name) -> Optional[TimeEvent]:
        raise NotImplementedError

    def __str__(self):
        return f"Repo name: {self.name}"


class TimeRepository(AbstractTimeRepository):
    def __init__(self, root_time_context: TimeContext):
        self.name = root_time_context.name
        self.root_time_context = root_time_context
        self.time_events: List[TimeEvent] = []
        self.time_contexts: Dict[str, TimeContext] = {}
        self.register(self.root_time_context)
        self.root_time_context.chrono.logger(f"Repository {self.name} created")

    def register(self, time_context: TimeContext):
        if not self.time_contexts.get(time_context.name, None):
            self.time_contexts[time_context.name] = time_context
            self.root_time_context.chrono.logger(f"Timer {time_context.name} registered in repo {self.name}")

    def add(self, time_event: TimeEvent):
        self.time_events.append(time_event)

    def get(self, time_context: TimeContext) -> List[TimeEvent]:
        return self.time_contexts.get(time_context.name).get_all()

    def get_all(self) -> List[TimeEvent]:
        return self.time_events

    def __str__(self):
        representation = ""
        for time_event in self.time_events:  # TODO Skip printing single tick events (e.g. create an enum to differentiate TimeEvent types)
            if isinstance(time_event, Label) or isinstance(time_event, Period):
                representation += str(time_event) + "\n"
        return representation


class RootTimeRepository(TimeRepository):

    def __init__(self, root_context: TimeContext):
        TimeRepository.__init__(self, root_context)


time_repo: Optional[TimeRepository] = None


def init_repo(time_context: TimeContext = None):
    """
    Initializes the root time_repo.
    """

    global time_repo
    if not time_repo:
        time_context.chrono.logger(f"Creating root repo {time_context.name}")
        time_repo = RootTimeRepository(time_context)
    return time_repo


def get_repo() -> Optional[TimeRepository]:
    """
    Returns the root time_repo.
    """
    return time_repo



import math
import time
import unittest.mock as mock

from chronologger import Timer
from chronologger.model import Chronologger, TimeUnit

sleep_time_seconds = 1

@mock.patch("chronologger.Timer")
def test_chronologger_logs_elapsed_time(mock_timer, capsys):
    # Basic timer functionality!
    timer = Chronologger("meh", mock)
    timer.start()
    time.sleep(sleep_time_seconds)  # Simulate some time has passed...
    period = timer.stop(reset=True)
    assert period.unit == TimeUnit.s
    assert math.isclose(period.time(), sleep_time_seconds, rel_tol=0.02)

@mock.patch("chronologger.Timer")
def test_chronologger_log_repoting(mock_timer, capsys):
    timer = Chronologger("meh", mock_timer)
    timer.start()
    timer.stop(reset=True)
    # assert False
    captured = capsys.readouterr()  # As default option in .stop() above does not log any message, then...
    assert "elapsed time" not in captured.out

    # This time log the standard message
    timer.start(start_tick_name="st")
    timer.stop(do_log=True, final_tick_name="ft")
    captured = capsys.readouterr()
    assert "elapsed time" in captured.out
    print(captured.out)

    # Check extended log messages
    timer = Chronologger('explicit_name', mock_timer, simple_log_msgs=False)
    assert timer.name == "explicit_name"
    timer.start(start_tick_name="st1")
    timer.stop(do_log=True, final_tick_name="ft1")
    captured = capsys.readouterr()
    assert "elapsed time" in captured.out
    assert "st1" in captured.out
    assert "ft1" in captured.out

# TODO: Differentiate (and separate) Timer from Logger tests
def test_as_context_manager(capsys):
    with Timer("test"):
        time.sleep(0)
    captured = capsys.readouterr()
    assert "test_start_tick" not in captured.out
    assert "test_end_tick" not in captured.out

    with Timer("test", log_when_exiting=True):
        time.sleep(0)
    captured = capsys.readouterr()
    assert "test_start_tick" in captured.out
    assert "test_end_tick" in captured.out

# TODO: Differentiate (and separate) Timer from Logger tests
def test_as_decorator(capsys):
    @Timer("test")
    def dummy():
        time.sleep(0)

    dummy()
    captured = capsys.readouterr()
    assert "test_start_tick" not in captured.out
    assert "test_end_tick" not in captured.out

    @Timer("test", log_when_exiting=True)
    def dummy():
        time.sleep(0)

    dummy()
    captured = capsys.readouterr()
    assert "test_start_tick" in captured.out
    assert "test_end_tick" in captured.out

# TODO: Differentiate (and separate) Timer from Logger tests
def test_unit_conversion(capsys):
    timer = Timer("test", unit=TimeUnit.ns).start()
    time.sleep(sleep_time_seconds)  # Simulate some time has passed...
    period = timer.stop(do_log=True)
    print(period)
    captured = capsys.readouterr()
    print(captured.out)
    print(sleep_time_seconds * 10 ** timer.chrono.unit.value)
    assert math.isclose(period.time(), sleep_time_seconds * 10 ** timer.chrono.unit.value, rel_tol=0.02)
    assert timer.chrono.unit.name in captured.out

# TODO: Differentiate (and separate) Timer from Logger tests
def test_marks(capsys):
    sleep_time_seconds = 0.1  # 100 ms
    timer = Timer(name="Test Loop!", unit=TimeUnit.ms).start("time to start loop")
    for i in range(3):
        time.sleep(sleep_time_seconds)
        timer.mark("i={}".format(i))
    period = timer.stop()
    print(timer.chrono.unit.from_secs(sleep_time_seconds * 3))
    assert math.isclose(period.elapsed(), timer.chrono.unit.from_secs(sleep_time_seconds * 3), rel_tol=0.1)

    print(timer)
    captured = capsys.readouterr()
    for i in range(3):
        assert "i={}".format(i) in captured.out

    with Timer(name="Test Loop in context!",
                      unit=TimeUnit.ms,
                      log_when_exiting=True,
                      simple_log=False) as timer:
        for j in range(3):
            time.sleep(sleep_time_seconds)
            timer.mark("j={}".format(j))

    captured = capsys.readouterr()
    for i in range(3):
        assert "j={}".format(i) in captured.out



from typing import Any, cast

from chronologger.model import Period, TimeContext, TimeEvent, Chronologger, TimeUnit, Label
from chronologger.repository import init_repo
from chronologger.service import record, register, show_time


class Timer(TimeContext):
    """A basic timer utility for logging time in code. It can be used as a class, context manager or decorator"""

    def __init__(self, name, unit=TimeUnit.s, simple_log=False, log_when_exiting=False, parent_ctx=None):
        self.name = name
        self.chrono = Chronologger(name, self, unit, simple_log_msgs=simple_log)
        self.log_when_exiting = log_when_exiting
        self.parent_ctx: TimeContext = parent_ctx
        self.chrono.logger(f"{self.name} Timer created!")

    def start(self, start_suffix: str = "_start_tick") -> "Timer":
        time_event: TimeEvent = self.chrono.start(self.name + start_suffix)
        record(time_event)
        return self

    def mark(self, name: str) -> "TimeEvent":
        time_event: TimeEvent = self.chrono.mark(name)
        record(time_event)
        return time_event

    def label(self, name: str):
        time_event: TimeEvent = Label(name, self)
        record(time_event)
        return time_event

    def stop(self, end_suffix: str = "_end_tick", do_log: bool = False, reset: bool = False) -> Period:
        time_event: Period = self.chrono.stop(self.name + end_suffix, do_log, reset)
        record(cast(TimeEvent, time_event))
        return time_event

    def print(self):
        show_time()

    """ContextDecorator implementation"""

    def __enter__(self) -> "Timer":
        """Start a new basic timer as a context manager"""
        register(self)
        self.start("_start_tick")
        return self

    """Presentation dunder implementations"""

    def __exit__(self, *exc_info: Any) -> None:
        """Stop the basic timer when exiting the context"""
        self.stop("_end_tick", do_log=self.log_when_exiting,
                  reset=True)  # TODO Make this last param configurable

    def __str__(self) -> str:
        if self.chrono.ticks.last() is None:
            tick_diff = ""
        else:
            tick_diff = (self.chrono.ticks.last() - self.chrono.ticks.first())
        representation = (
            f"\n------------------------------------------------------------------------------------------\n"
            f"{self.chrono.name}\t{tick_diff}\n"
            f"{self.chrono.description}\n"
            f"Has parent context: {self.parent_ctx}\n"
            f"------------------------------------------------------------------------------------------\n"
        )
        if len(self.chrono.ticks) > 2:
            representation += (
                f"Marks:\n"
                f"{self.chrono.ticks}\n"
                f"------------------------------------------------------------------------------------------\n"
            )
        return representation


root_timer = Timer(name="root")
root_repo = init_repo(root_timer)



import math
import time
import unittest.mock as mock

from chronologger.model import Tick, TimeUnit, EventRecorder, TimeEvent, Period

time_event_name = "tick"


@mock.patch("chronologger.Timer")
def test_tick_basic_properties(mock_timer):
    tick = Tick(time_event_name, mock_timer)
    assert isinstance(tick, TimeEvent)
    assert isinstance(tick, Tick)
    assert tick.name == time_event_name
    assert tick.unit == TimeUnit.s
    now = time.perf_counter()
    assert tick.time() < now


@mock.patch("chronologger.Timer")
def test_tick_conversions(mock_timer):
    tick = Tick(time_event_name, mock_timer)

    tick_in_secs_again = tick.to(TimeUnit.s)
    assert tick_in_secs_again == tick

    tick_ms = tick.to(TimeUnit.ms)
    assert tick_ms.name == tick.name
    assert tick_ms.unit == TimeUnit.ms
    assert math.isclose(tick_ms.time(), TimeUnit.ms.from_secs(tick.time()), rel_tol=0.02)

    tick_ns = tick.to(TimeUnit.ns)
    assert tick_ns.name == tick.name
    assert tick_ns.unit == TimeUnit.ns
    assert math.isclose(tick_ns.time(), TimeUnit.ns.from_secs(tick.time()), rel_tol=0.02)
    now_in_ns = TimeUnit.ns.from_secs(time.perf_counter())
    assert tick_ns.time() < now_in_ns


@mock.patch("chronologger.Timer")
def test_two_ticks_with_the_same_name_are_different_as_they_have_different_timestamps(mock_timer):
    tick_1 = Tick("tock", mock_timer)
    tick_2 = Tick("tock", mock_timer)
    assert not tick_1 == tick_2
    assert tick_1.time() < tick_2.time()


@mock.patch("chronologger.Timer")
def test_basic_homogeneus_period(mock_timer):
    sleep_time_secs = 0.1  # 100 ms
    tick_1 = Tick("tick", mock_timer)
    time.sleep(sleep_time_secs)
    tick_2 = Tick("tock", mock_timer)

    explicit_period = Period("test_period", mock_timer, TimeUnit.s, tick_1, tick_2)
    assert isinstance(explicit_period, TimeEvent)
    assert isinstance(explicit_period, Period)
    assert explicit_period.name == "test_period"
    assert explicit_period.unit == TimeUnit.s
    assert math.isclose(explicit_period.elapsed(), sleep_time_secs, rel_tol=0.05)

    calculated_period = tick_2 - tick_1
    assert isinstance(calculated_period, TimeEvent)
    assert isinstance(calculated_period, Period)
    assert calculated_period.name == "elapsed"  # This is the default name
    assert calculated_period.unit == TimeUnit.s
    assert math.isclose(calculated_period.elapsed(), sleep_time_secs, rel_tol=0.05)


@mock.patch("chronologger.Timer")
def test_basic_heterogeneous_period(mock_timer, capsys):
    sleep_time_secs = 0.1  # 100 ms
    tick_1 = Tick("tick", mock_timer)
    time.sleep(sleep_time_secs)
    tick_2 = Tick("tock", mock_timer, TimeUnit.ns)

    period_name = "test_period"
    explicit_period_ms = Period(period_name, mock_timer, TimeUnit.ms, tick_1, tick_2)
    tick_1_ms = tick_1.to(TimeUnit.ms).time()
    tick_2_ms = tick_2.to(TimeUnit.ms).time()
    print(tick_1_ms)
    print(tick_2_ms)
    print(explicit_period_ms.elapsed())
    assert math.isclose(explicit_period_ms.elapsed(), tick_2_ms - tick_1_ms, abs_tol=0.05)
    print(explicit_period_ms)
    captured = capsys.readouterr()
    assert period_name in captured.out
    assert TimeUnit.ms.name in captured.out


@mock.patch("chronologger.Timer")
def test_heterogeneous_period_gets_converted_properly_to_homogeneous(mock_timer, capsys):
    sleep_time_secs = 0.1  # 100 ms
    tick_1 = Tick("tick", mock_timer)
    time.sleep(sleep_time_secs)
    tick_2 = Tick("tock", mock_timer, TimeUnit.ns)

    heterogeneous_period = tick_2 - tick_1
    homogeneous_period: Period = heterogeneous_period.to(TimeUnit.ms)
    assert homogeneous_period.name == "elapsed"
    assert homogeneous_period.start == tick_1
    assert homogeneous_period.end == tick_2
    assert homogeneous_period.start.unit == TimeUnit.s
    assert homogeneous_period.end.unit == TimeUnit.ns
    assert math.isclose(homogeneous_period.elapsed(),
                        homogeneous_period.end.to(TimeUnit.ms).time() - homogeneous_period.start.to(TimeUnit.ms).time(),
                        abs_tol=0.05)

    print(homogeneous_period)
    captured = capsys.readouterr()
    assert "elapsed" in captured.out
    assert TimeUnit.ms.name in captured.out


# TODO The next four tests can be combined using text fixtures and validation data passed as arguments
@mock.patch("chronologger.Timer")
def test_homogeneous_explicit_period_creation(mock_timer, capsys):
    """This should test that the conversions and calculations are correct
    given that the main time unit driving the process is the one that
    corresponds to the first period. As all the ticks are explicitly passed,
    no conversions are needed"""

    driver_time_unit = TimeUnit.s

    sleep_time_secs = 0.1  # 100 ms
    tick_1 = Tick("tick", mock_timer)
    time.sleep(sleep_time_secs)
    tick_2 = Tick("tack", mock_timer)

    explicit_period_1 = Period("period_1", mock_timer, driver_time_unit, tick_1, tick_2)

    sleep_time_secs = 0.1  # 100 ms
    tick_3 = Tick("tock", mock_timer)
    time.sleep(sleep_time_secs)
    tick_4 = Tick("tuck", mock_timer)

    explicit_period_2 = Period("period_2", mock_timer, TimeUnit.s, tick_3, tick_4)

    combined_explicit_period = Period("new_period", mock_timer, TimeUnit.s, explicit_period_1.start,
                                      explicit_period_2.end)

    assert combined_explicit_period.name == "new_period"
    assert combined_explicit_period.unit == driver_time_unit
    assert combined_explicit_period.start == tick_1
    assert combined_explicit_period.end == tick_4


@mock.patch("chronologger.Timer")
def test_homogeneous_calculated_period_difference(mock_timer, capsys):
    """This should test that the conversions and calculations are correct
    given that the main time unit driving the process is the one that
    corresponds to the first period. As all the ticks involved are
    homogeneous with regard to the time units, no conversions are needed"""

    driver_time_unit = TimeUnit.s

    sleep_time_secs = 0.1  # 100 ms
    tick_1 = Tick("tick", mock_timer)
    time.sleep(sleep_time_secs)
    tick_2 = Tick("tack", mock_timer)

    explicit_period_1 = Period("period_1", mock_timer, driver_time_unit, tick_1, tick_2)

    sleep_time_secs = 0.1  # 100 ms
    tick_3 = Tick("tock", mock_timer)
    time.sleep(sleep_time_secs)
    tick_4 = Tick("tuck", mock_timer)

    explicit_period_2 = Period("period_2", mock_timer, TimeUnit.s, tick_3, tick_4)

    combined_calculated_period = explicit_period_2 - explicit_period_1

    assert combined_calculated_period.name == f"elapsed ({tick_4.name} - {tick_1.name})"
    assert combined_calculated_period.unit == driver_time_unit
    assert combined_calculated_period.start == tick_1
    assert combined_calculated_period.end == tick_4


@mock.patch("chronologger.Timer")
def test_heterogeneous_explicit_period_creation(mock_timer, capsys):
    """This should test that the conversions and calculations are correct
    given that the main time unit driving the process is the one that
    corresponds to the first period. As all the ticks involved are
    homogeneous with regard to the time units, no conversions are needed"""

    driver_time_unit = TimeUnit.s

    sleep_time_secs = 0.1  # 100 ms
    tick_1 = Tick("tick", mock_timer, TimeUnit.ns)
    time.sleep(sleep_time_secs)
    tick_2 = Tick("tack", mock_timer, TimeUnit.ms)

    explicit_period_1 = Period("period_1", mock_timer, driver_time_unit, tick_1, tick_2)

    sleep_time_secs = 0.1  # 100 ms
    tick_3 = Tick("tock", mock_timer, TimeUnit.ns)
    time.sleep(sleep_time_secs)
    tick_4 = Tick("tuck", mock_timer, TimeUnit.ms)

    explicit_period_2 = Period("period_2", mock_timer, TimeUnit.ms, tick_3, tick_4)

    combined_calculated_period = Period("new_period", mock_timer, driver_time_unit, explicit_period_1.start,
                                        explicit_period_2.end)

    assert combined_calculated_period.name == f"new_period"
    assert combined_calculated_period.unit == driver_time_unit
    assert combined_calculated_period.start.name == tick_1.name
    assert combined_calculated_period.start.unit == tick_1.unit
    assert combined_calculated_period.end.name == tick_4.name
    assert combined_calculated_period.end.unit == tick_4.unit


@mock.patch("chronologger.Timer")
def test_heterogeneous_calculated_period_difference(mock_timer, capsys):
    """This should test that the conversions and calculations are correct
    given that the main time unit driving the process is the one that
    corresponds to the first period. Conversions of the ticks involved
    are expected"""

    driver_time_unit = TimeUnit.s

    sleep_time_secs = 0.1  # 100 ms
    tick_1 = Tick("tick", mock_timer, TimeUnit.ns)
    time.sleep(sleep_time_secs)
    tick_2 = Tick("tack", mock_timer, TimeUnit.ms)

    explicit_period_1 = Period("period_1", mock_timer, driver_time_unit, tick_1, tick_2)

    sleep_time_secs = 0.1  # 100 ms
    tick_3 = Tick("tock", mock_timer, TimeUnit.ns)
    time.sleep(sleep_time_secs)
    tick_4 = Tick("tuck", mock_timer, TimeUnit.ms)

    explicit_period_2 = Period("period_2", mock_timer, TimeUnit.ms, tick_3, tick_4)

    combined_calculated_period = explicit_period_2 - explicit_period_1

    assert combined_calculated_period.name == f"elapsed ({tick_4.name} - {tick_1.name})"
    assert combined_calculated_period.unit == driver_time_unit
    assert combined_calculated_period.start.name == tick_1.name
    assert combined_calculated_period.start.unit == driver_time_unit
    assert combined_calculated_period.end.name == tick_4.name
    assert combined_calculated_period.end.unit == driver_time_unit


@mock.patch("chronologger.Timer")
def test_event_recorder_can_hold_and_retrieve_events(mock_timer):
    num_of_time_events_to_create = 3
    event_recorder = EventRecorder()
    for i in range(0, num_of_time_events_to_create):
        event_recorder.add(Tick(f"evt {i}", mock_timer))
        time.sleep(0.1)
    assert len(event_recorder.get_all()) == num_of_time_events_to_create


@mock.patch("chronologger.Timer")
def test_event_recorder_can_homogeinize_different_event_time_units(mock_timer):
    event_recorder = EventRecorder()
    tick_1 = Tick(f"evt 1", mock_timer, unit=TimeUnit.s)
    event_recorder.add(tick_1)
    time.sleep(0.1)
    tick_2 = Tick(f"evt 2", mock_timer, unit=TimeUnit.ns)
    event_recorder.add(tick_2)
    assert len(event_recorder.get_all()) == 2

    converted_event_recorder: EventRecorder = event_recorder.to(TimeUnit.ms)
    for event in converted_event_recorder.get_all():
        assert event.unit == TimeUnit.ms



from chronologger import Tick, Timer, TimeUnit
from chronologger.timer import root_repo
from chronologger.repository import TimeRepository, RootTimeRepository


def test_root_repo_is_initialized():
    assert isinstance(root_repo, RootTimeRepository)
    assert root_repo.name == 'root'


def test_nodes_at_root_level():
    root_timer = Timer("root")
    repo = TimeRepository(root_timer)

    t1 = Tick("phase 1", root_timer)
    t2 = Tick("phase 2", root_timer)
    t3 = Tick("phase 3", root_timer)

    repo.add(t1)
    repo.add(t2)
    repo.add(t3)

    events = repo.get_all()
    assert events == list([t1, t2, t3])


def test_nested_nodes_1_level():
    root_timer = Timer("root")
    repo = TimeRepository(root_timer)

    t1 = Tick("tick 1", root_timer)
    repo.add(t1)

    events = repo.get_all()
    assert events == list([t1])

    p1_timer = Timer("phase_1", unit=TimeUnit.ms, parent_ctx=root_timer)
    repo.register(p1_timer)

    repo.add(Tick("simulated tick 1.1", p1_timer))
    repo.add(Tick("simulated tick 1.2", p1_timer))

    events = repo.get_all()
    assert len(
        events) == 3  # TODO For now it's a flat structure and we can't check nested timer contents as we have to invoke the service

    assert len(repo.time_contexts) == 2



from .model import TimeEvent, Label
from .repository import get_repo
from .timer import TimeContext


def register(timer: TimeContext):
    get_repo().register(timer)


def label(name: str):
    time_event = Label(name, None)
    record(time_event)


def record(time_event: TimeEvent):
    """Records every time event happening in the system"""
    get_repo().add(time_event)


def show_time():
    """It's show time!!!"""
    print(get_repo())



from typing import List

from chronologger.model import Tick, TimeEvent
from chronologger.repository import AbstractTimeRepository
from chronologger.service import record

time_event_name = "tick"

#
# class FlatTimeRepository(AbstractTimeRepository):
#     time_events: List[TimeEvent] = list()
#
#     def __init__(self, name):
#         self.name = name
#
#     def add(self, time_event: TimeEvent):
#         self.time_event.append(time_event)
#
#     def get(self):
#         self.time_events[-1]
#
#     def add(self, TimeUnit):
#         pass


# def test_events_recording():
#     tick = Tick(time_event_name)
#     last_recorded_time_event = record(tick)
#     assert last_recorded_time_event == tick



from dataclasses import dataclass, field


@dataclass(frozen=True)
class Core:
    surname: str = field(default="Adios")


@dataclass(frozen=True)
class Extension(Core):
    name: str = field(default="Hola")



import enum
import itertools
import time
from abc import abstractmethod, ABC
from contextlib import ContextDecorator
from dataclasses import dataclass, field, replace
from typing import List, Optional, ClassVar, Callable, Any

try:
    from typing import Protocol, runtime_checkable
except ImportError:
    from typing_extensions import Protocol, runtime_checkable


class ChronologgerError(Exception):
    """A general exception used to report errors in use of the Chronologger project"""


class TimeUnit(enum.Enum):
    ns = (9)
    ms = (3)
    s = (0)

    def __init__(self, x) -> None:
        self.x = x

    def from_secs(self, secs: float) -> float:
        return secs * 10 ** self.x

    def to_secs(self, timelapse: float) -> float:
        return timelapse / 10 ** self.x


@runtime_checkable
class TimeEvent(Protocol):
    """Main concept of the library, representing a discrete time event"""
    name: str
    time_context: 'TimeContext'
    unit: TimeUnit

    @abstractmethod
    def time(self) -> float:
        raise NotImplementedError

    @abstractmethod
    def to(self, unit: TimeUnit) -> 'TimeEvent':
        raise NotImplementedError

    @abstractmethod
    def __sub__(self, other: 'TimeEvent') -> 'Period':
        raise NotImplementedError


@dataclass(frozen=True)
class Tick:
    """A discrete time event implementation"""
    name: str
    time_context: 'TimeContext'
    unit: TimeUnit = TimeUnit.s
    # I decided to store the time in secs but this is just an implementation detail
    _tick_in_secs: float = field(default_factory=lambda: time.perf_counter(), init=False, repr=False)

    def time(self) -> float:
        """Returns the value of the time in the TimeUnits in which the object is specified"""
        return self.unit.from_secs(self._tick_in_secs)

    def to(self, unit: TimeUnit) -> 'Tick':
        if unit == self.unit:
            return self
        else:  # Create new object with a new time unit. Immutability broken, but just in the creation of a new object
            new_object = replace(self, unit=unit)
            object.__setattr__(new_object, '_tick_in_secs', self._tick_in_secs)
            return new_object

    def __sub__(self, other: TimeEvent) -> 'Period':
        if other is None:
            return self
        return Period("elapsed", self.time_context, other.unit, other, self)

    def __str__(self) -> str:
        return f"{self.name}: {self.time():.3f} {self.unit.name}"


@dataclass(frozen=True)
class Label(Tick):
    fillup_char: chr = "*"
    repeat_no: int = 80

    def __str__(self) -> str:
        return (f"{self.fillup_char * self.repeat_no}\n"
                f"{self.name}: {self.time():.3f} {self.unit.name}\n"
                f"{self.fillup_char * self.repeat_no}\n")


@dataclass(frozen=True)
class Period:
    """Represents the elapsed time between two time events.

    We allow heterogeneous Periods when explicitly created, meaning
    that the ticks passed can have different TimeUnits. Calculations to
    reconcile the results are done when results are presented."
    """
    name: str
    time_context: 'TimeContext'
    unit: TimeUnit
    start: TimeEvent
    end: TimeEvent

    # If we don't want to allow heterogeneus Periods (see class description) uncomment
    # this and change the semantics
    # def __post_init__(self):
    #     if self.start.unit != self.unit or self.end.unit != self.unit:
    #         message = (
    #             f"Ether start or end ticks have different units from Period units."
    #             f"start tick {self.start.name} -> {self.start.unit.name}"
    #             f"end tick {self.end.name} -> {self.end.unit.name}"
    #             f"Please convert them first"
    #         )
    #         raise ChronologgerError(message)

    def elapsed(self) -> float:
        """Returns the value of the elapsed time in the TimeUnits in which the object is specified"""
        start_in_current_units = self.start.to(self.unit).time()
        end_in_current_units = self.end.to(self.unit).time()
        return end_in_current_units - start_in_current_units

    def time(self) -> float:
        """Returns the value of the elapsed time in the TimeUnits in which the object is specified"""
        return self.elapsed()

    def to(self, unit: TimeUnit) -> 'Period':
        return Period(self.name, self.time_context, unit, self.start, self.end)

    def __sub__(self, other: 'Period') -> 'Period':
        """Builds a new Period from the parts of the two periods involved.

        Instead of keeping a Frankenstein period, we get the start tick from
        the subtracting period and the end tick from the minuend. The time
        unit from the new period will be the one from the start_tick."""
        new_time_unit = other.unit
        new_start_tick = other.start.to(new_time_unit)
        new_end_tick = self.end.to(new_time_unit)
        return Period(f"elapsed ({new_end_tick.name} - {new_start_tick.name})",
                      self.time_context, new_time_unit, new_start_tick, new_end_tick)

    def __str__(self) -> str:
        return f"{self.time():.3f} {self.unit.name} ({self.name})   =    {self.end} - {self.start}"


@dataclass(frozen=True)
class EventRecorder:
    """Stores a list of time events"""
    events: List[TimeEvent] = field(default_factory=list)

    def add(self, event: TimeEvent) -> Optional[Period]:
        """Add a new time event to the recorded list

        Parameters
        ----------
        event : The time event to store

        Returns
        -------
        the (optional) period (TimeEvent) from the initial time event recorded
        """
        self.events.append(event)
        return self.elapsed()

    def first(self) -> Optional[TimeEvent]:
        return None if len(self.events) == 0 else self.events[0]

    def last(self) -> Optional[TimeEvent]:
        return None if len(self.events) == 0 else self.events[-1]

    def elapsed(self) -> Optional[Period]:
        return None if len(self.events) == 0 else self.last() - self.first()

    def get_all(self) -> List[TimeEvent]:
        return self.events

    def to(self, unit: TimeUnit):
        converted_event_recorder = EventRecorder()
        for event in self.events:
            converted_event_recorder.add(event.to(unit))
        return converted_event_recorder

    def __len__(self) -> int:
        return len(self.events)

    def __str__(self) -> str:
        if len(self.events) == 0:
            return ""
        previous_marker = self.events[0]
        time_unit = previous_marker.unit  # Reporting in the time unit of the first TimeEvent
        markers_str = ""
        i = 1
        while i < len(self.events):
            marker = self.events[i]
            period = Period(marker.name, previous_marker.time_context, time_unit, previous_marker, marker)
            markers_str += f"\t- {period}\n"
            previous_marker = marker
            i += 1
        return markers_str


def event_recorder():
    return EventRecorder()


@dataclass(frozen=False)
class Chronologger:
    """Core class. Manages the creation of every tick and period."""

    id_iter: ClassVar[int] = itertools.count()

    name: Optional[str]
    time_context: 'TimeContext'
    unit: TimeUnit = TimeUnit.s
    description: str = ""
    logger: Callable[[str], None] = print
    simple_log_msgs: bool = True
    ticks: EventRecorder = field(default_factory=event_recorder)

    parent: Optional[str] = None

    def __post_init__(self) -> None:  # TODO This is not necessary anymore... but leave it for now...
        """Initialization: add unique name at least"""
        if not self.name:
            object.__setattr__(self, 'name', "timer_" + str(next(Chronologger.id_iter)))

    def start(self, start_tick_name: str = "start_tick") -> TimeEvent:
        """Start a new basic timer"""
        time_event = Tick(start_tick_name, self.time_context, self.unit)
        self.ticks.add(time_event)
        return time_event

    def _report_time(self, do_log, period):
        if self.logger and do_log:
            self.logger(f"{period.time():3f} {period.unit.name} elapsed time") if self.simple_log_msgs else self.logger(
                self)

    def stop(self, final_tick_name: str = "end_tick", do_log: bool = False, reset: bool = False) -> Period:
        """Stop the basic timer reporting the elapsed time"""
        if len(self.ticks) == 0:
            raise ChronologgerError(f"Timer not started yet! Use .start() to start counting time...")

        tick = Tick(final_tick_name, self.time_context, unit=self.unit)

        period: Period = self.ticks.add(tick)
        self._report_time(do_log, period)

        self.reset() if reset else None

        return period.to(self.unit)

    def mark(self, name: str) -> TimeEvent:
        tick = Tick(name, self.time_context, unit=self.unit)
        self.ticks.add(tick)
        return tick

    def reset(self) -> None:
        self.ticks = EventRecorder()

    def get_all(self) -> List[TimeEvent]:
        return self.ticks.get_all()


class TimeContext(ContextDecorator, ABC):
    name: str
    parent_ctx: "TimeContext"
    decorated: bool
    chrono: Chronologger = field(default_factory=Chronologger)
    log_when_exiting: bool = False

    # TODO Create method to allow to report phases e.g. def phase():

    def get_all(self):
        return self.chrono.get_all()

    @abstractmethod
    def __enter__(self) -> "TimeContext":
        pass

    @abstractmethod
    def __exit__(self, *exc_info: Any):
        pass
